
import enum


class Intents(enum.IntFlag):
    '''Intents of the user.'''

    GUILDS = (1 << 0)
    # GUILD_CREATE = (1 << 0)
    # GUILD_UPDATE = (1 << 0)
    # GUILD_DELETE = (1 << 0)
    # GUILD_ROLE_CREATE = (1 << 0)
    # GUILD_ROLE_UPDATE = (1 << 0)
    # GUILD_ROLE_DELETE = (1 << 0)
    # CHANNEL_CREATE = (1 << 0)
    # CHANNEL_UPDATE = (1 << 0)
    # CHANNEL_DELETE = (1 << 0)
    # CHANNEL_PINS_UPDATE = (1 << 0)
    # THREAD_CREATE = (1 << 0)
    # THREAD_UPDATE = (1 << 0)
    # THREAD_DELETE = (1 << 0)
    # THREAD_LIST_SYNC = (1 << 0)
    # THREAD_MEMBER_UPDATE = (1 << 0)
    # THREAD_MEMBERS_UPDATE = (1 << 0)
    # STAGE_INSTANCE_CREATE = (1 << 0)
    # STAGE_INSTANCE_UPDATE = (1 << 0)
    # STAGE_INSTANCE_DELETE = (1 << 0)

    GUILD_MEMBERS = (1 << 1)
    # GUILD_MEMBER_ADD = (1 << 1)
    # GUILD_MEMBER_UPDATE = (1 << 1)
    # GUILD_MEMBER_REMOVE = (1 << 1)
    # THREAD_MEMBERS_UPDATE = (1 << 1)

    GUILD_BANS = (1 << 2)
    # GUILD_BAN_ADD = (1 << 2)
    # GUILD_BAN_REMOVE = (1 << 2)

    GUILD_EMOJIS_AND_STICKERS = (1 << 3)
    # GUILD_EMOJIS_UPDATE = (1 << 3)
    # GUILD_STICKERS_UPDATE = (1 << 3)

    GUILD_INTEGRATIONS = (1 << 4)
    # GUILD_INTEGRATIONS_UPDATE = (1 << 4)
    # INTEGRATION_CREATE = (1 << 4)
    # INTEGRATION_UPDATE = (1 << 4)
    # INTEGRATION_DELETE = (1 << 4)

    GUILD_WEBHOOKS = (1 << 5)
    # WEBHOOKS_UPDATE = (1 << 5)

    GUILD_INVITES = (1 << 6)
    # INVITE_CREATE = (1 << 6)
    # INVITE_DELETE = (1 << 6)

    GUILD_VOICE_STATES = (1 << 7)
    # VOICE_STATE_UPDATE = (1 << 7)

    GUILD_PRESENCES = (1 << 8)
    # PRESENCE_UPDATE = (1 << 8)

    GUILD_MESSAGES = (1 << 9)
    # MESSAGE_CREATE = (1 << 9)
    # MESSAGE_UPDATE = (1 << 9)
    # MESSAGE_DELETE = (1 << 9)
    # MESSAGE_DELETE_BULK = (1 << 9)

    GUILD_MESSAGE_REACTIONS = (1 << 10)
    # MESSAGE_REACTION_ADD = (1 << 10)
    # MESSAGE_REACTION_REMOVE = (1 << 10)
    # MESSAGE_REACTION_REMOVE_ALL = (1 << 10)
    # MESSAGE_REACTION_REMOVE_EMOJI = (1 << 10)

    GUILD_MESSAGE_TYPING = (1 << 11)
    # TYPING_START = (1 << 11)

    DIRECT_MESSAGES = (1 << 12)
    # MESSAGE_CREATE = (1 << 12)
    # MESSAGE_UPDATE = (1 << 12)
    # MESSAGE_DELETE = (1 << 12)
    # CHANNEL_PINS_UPDATE = (1 << 12)

    DIRECT_MESSAGE_REACTIONS = (1 << 13)
    # MESSAGE_REACTION_ADD = (1 << 13)
    # MESSAGE_REACTION_REMOVE = (1 << 13)
    # MESSAGE_REACTION_REMOVE_ALL = (1 << 13)
    # MESSAGE_REACTION_REMOVE_EMOJI = (1 << 13)

    DIRECT_MESSAGE_TYPING = (1 << 14)
    # TYPING_START = (1 << 14)



import warnings
import inspect
import functools
from cachetools import TTLCache
from dataclasses import dataclass
from datetime import datetime, timedelta
from typing import Callable, Dict, Optional

from ..objects import interactions, snowflake
from ..utilities import Log


@dataclass
class CallbackData:
    '''Callback dataclass.'''
    custom_id: str
    callback: Callable
    unlimited: bool


class CommandHandler:
    '''Manage commands for the user.'''
    _log = Log()

    registered_commands: Dict['snowflake.Snowflake', Callable] = dict()

    global_lookup: Dict[str, Callable] = dict()

    guild_lookup: Dict[str, Callable] = dict()

    registered_custom_ids: 'TTLCache[str, CallbackData]' = TTLCache(
        maxsize=float('inf'),
        ttl=timedelta(minutes=15),  # type: ignore
        timer=datetime.now,  # type: ignore
    )

    @classmethod
    def register_global_callback(cls, command_name, callback_function, command_id: Optional['snowflake.Snowflake'] = None) -> None:
        '''Register name of a command to a given callback function globally.'''
        cls.global_lookup[command_name] = callback_function
        if command_id is not None:
            cls.registered_commands[command_id] = callback_function

    @classmethod
    def register_guild_callback(cls, command_name, callback_function, command_id: Optional['snowflake.Snowflake'] = None) -> None:
        '''Register name of a command to a given callback function for guilds.'''
        cls.guild_lookup[command_name] = callback_function
        if command_id is not None:
            cls.registered_commands[command_id] = callback_function

    @classmethod
    def register_interaction_custom_id(cls, custom_id: str, callback_function: Callable, unlimited: bool = False):
        '''Register a callback for an interaction.

        Arguments:
            custom_id (str): The custom id to trigger on.
            callback_function (Callable): Function to call when interaction is triggered. MUST be asnyc.
            unlimited (bool): Allows more than one interaction to be called. The user is expected to deregister this interaction later. If
                no cleanup occurs, the handler is cleaned after 15 minutes.
        '''
        # TODO: We can probably customize the duration of the cache at runtime?
        obj = CallbackData(custom_id, callback_function, unlimited)
        cls.registered_custom_ids[custom_id] = obj

    @classmethod
    def unregister_interaction_custom_id(cls, custom_id: str, not_exists_ok: bool = False):
        '''Attempt to remove a custom_id from the interaction cache. Warn if already removed.'''
        if custom_id in cls.registered_custom_ids:
            del cls.registered_custom_ids[custom_id]
        elif not not_exists_ok:
            cls._log.warning(f'Observed improper attempt to remove interaction_id [{custom_id}].')

    @classmethod
    async def command_handler(cls, client, interaction: 'interactions.InteractionStructure') -> None:  # noqa: C901
        '''Handle incoming commands and dispatch them to the correct type handler.'''
        if interaction.data is None:
            return

        if interaction.type == interactions.enumerations.INTERACTION_TYPES.MESSAGE_COMPONENT:
            await cls.handle_message_component(client, interaction)
        elif interaction.type == interactions.enumerations.INTERACTION_TYPES.APPLICATION_COMMAND:
            await cls.handle_application_command(client, interaction)

    @classmethod
    async def handle_application_command(cls, client, interaction: 'interactions.InteractionStructure') -> None:
        '''Handle interactions against Messages and Users.'''
        assert interaction.data is not None

        if interaction.data.id in cls.registered_commands:
            await cls.registered_commands[interaction.data.id](client, interaction)
            return

        if interaction.data.id not in cls.registered_commands:
            # Lookup command in global
            results = None
            try:
                results = await client.API.get_global_application_command(interaction.data.id)
            except Exception:
                pass
            if (results is not None) and (results['name'] in cls.global_lookup):
                cls.registered_commands[interaction.data.id] = cls.global_lookup[results['name']]
                await cls.registered_commands[interaction.data.id](client, interaction)
                return
            # Lookup command in guild
            if not hasattr(interaction, 'guild_id'):
                return
            try:
                assert type(interaction.guild_id) is snowflake.Snowflake
                results = await client.API.get_guild_application_command(interaction.guild_id, interaction.data.id)
            except Exception:
                pass
            if (results is not None) and (results['name'] in cls.guild_lookup):
                cls.registered_commands[interaction.data.id] = cls.guild_lookup[results['name']]
                await cls.registered_commands[interaction.data.id](client, interaction)
                return
            raise LookupError(f'Unable to find interaction {interaction.data.id} in Global or Guild!')

    @classmethod
    async def handle_message_component(cls, client, interaction: 'interactions.InteractionStructure') -> None:
        '''Handle slash-commands.'''
        assert interaction.data is not None
        cls._log.info(f'Saw id [{interaction.id}] with custom id [{interaction.data.custom_id}].')
        assert interaction.data is not None
        if interaction.data.custom_id not in cls.registered_custom_ids:
            raise RuntimeError('Got unexpected interaction')

        callback_data = cls.registered_custom_ids[interaction.data.custom_id]
        if callback_data.unlimited is False:
            del cls.registered_custom_ids[interaction.data.custom_id]

        if cls.__iscoroutinefunction_or_partial(callback_data.callback):
            await callback_data.callback(client, interaction)
        else:
            warnings.warn('While sync functions are supported as callbacks in CommandHandler, they are STRONGLY counter-recommended!', UserWarning)
            callback_data.callback(client, interaction)

    @classmethod
    def __iscoroutinefunction_or_partial(cls, object):
        while isinstance(object, functools.partial):
            object = object.func
        return inspect.iscoroutinefunction(object)



import enum


class DISCORD_EVENTS(enum.Enum):
    '''Enumeration of various discord events.

    Attributes:
        CHANNEL_CREATE (enum):
        CHANNEL_DELETE (enum):
        CHANNEL_PINS_UPDATE (enum):
        CHANNEL_UPDATE (enum):
        GUILD_BAN_ADD (enum):
        GUILD_BAN_REMOVE (enum):
        GUILD_CREATE (enum):
        GUILD_DELETE (enum):
        GUILD_EMOJIS_UPDATE (enum):
        GUILD_INTEGRATIONS_UPDATE (enum):
        GUILD_MEMBER_ADD (enum):
        GUILD_MEMBER_REMOVE (enum):
        GUILD_MEMBER_UPDATE (enum):
        GUILD_ROLE_CREATE (enum):
        GUILD_ROLE_DELETE (enum):
        GUILD_ROLE_UPDATE (enum):
        GUILD_STICKERS_UPDATE (enum):
        GUILD_UPDATE (enum):
        INTEGRATION_CREATE (enum):
        INTEGRATION_DELETE (enum):
        INTEGRATION_UPDATE (enum):
        INVITE_CREATE (enum):
        INVITE_DELETE (enum):
        MESSAGE_CREATE (enum):
        MESSAGE_DELETE (enum):
        MESSAGE_DELETE_BULK (enum):
        MESSAGE_REACTION_ADD (enum):
        MESSAGE_REACTION_REMOVE (enum):
        MESSAGE_REACTION_REMOVE_ALL (enum):
        MESSAGE_REACTION_REMOVE_EMOJI (enum):
        MESSAGE_UPDATE (enum):
        PRESENCE_UPDATE (enum):
        READY (enum):
        STAGE_INSTANCE_CREATE (enum):
        STAGE_INSTANCE_DELETE (enum):
        STAGE_INSTANCE_UPDATE (enum):
        THREAD_CREATE (enum):
        THREAD_DELETE (enum):
        THREAD_LIST_SYNC (enum):
        THREAD_MEMBER_UPDATE (enum):
        THREAD_MEMBERS_UPDATE (enum):
        THREAD_UPDATE (enum):
        TYPING_START (enum):
        VOICE_STATE_UPDATE (enum):
        WEBHOOKS_UPDATE (enum):
        INTERACTION_CREATE (enum):
    '''
    CHANNEL_CREATE = enum.auto()
    CHANNEL_DELETE = enum.auto()
    CHANNEL_PINS_UPDATE = enum.auto()
    CHANNEL_UPDATE = enum.auto()
    GUILD_BAN_ADD = enum.auto()
    GUILD_BAN_REMOVE = enum.auto()
    GUILD_CREATE = enum.auto()
    GUILD_DELETE = enum.auto()
    GUILD_EMOJIS_UPDATE = enum.auto()
    GUILD_INTEGRATIONS_UPDATE = enum.auto()
    GUILD_MEMBER_ADD = enum.auto()
    GUILD_MEMBER_REMOVE = enum.auto()
    GUILD_MEMBER_UPDATE = enum.auto()
    GUILD_ROLE_CREATE = enum.auto()
    GUILD_ROLE_DELETE = enum.auto()
    GUILD_ROLE_UPDATE = enum.auto()
    GUILD_STICKERS_UPDATE = enum.auto()
    GUILD_UPDATE = enum.auto()
    INTEGRATION_CREATE = enum.auto()
    INTEGRATION_DELETE = enum.auto()
    INTEGRATION_UPDATE = enum.auto()
    INVITE_CREATE = enum.auto()
    INVITE_DELETE = enum.auto()
    MESSAGE_CREATE = enum.auto()
    MESSAGE_DELETE = enum.auto()
    MESSAGE_DELETE_BULK = enum.auto()
    MESSAGE_REACTION_ADD = enum.auto()
    MESSAGE_REACTION_REMOVE = enum.auto()
    MESSAGE_REACTION_REMOVE_ALL = enum.auto()
    MESSAGE_REACTION_REMOVE_EMOJI = enum.auto()
    MESSAGE_UPDATE = enum.auto()
    PRESENCE_UPDATE = enum.auto()
    READY = enum.auto()
    STAGE_INSTANCE_CREATE = enum.auto()
    STAGE_INSTANCE_DELETE = enum.auto()
    STAGE_INSTANCE_UPDATE = enum.auto()
    THREAD_CREATE = enum.auto()
    THREAD_DELETE = enum.auto()
    THREAD_LIST_SYNC = enum.auto()
    THREAD_MEMBER_UPDATE = enum.auto()
    THREAD_MEMBERS_UPDATE = enum.auto()
    THREAD_UPDATE = enum.auto()
    TYPING_START = enum.auto()
    VOICE_STATE_UPDATE = enum.auto()
    WEBHOOKS_UPDATE = enum.auto()
    INTERACTION_CREATE = enum.auto()


class MESSAGE_TYPE(enum.IntEnum):
    '''Types of messages.'''
    DEFAULT = 0
    RECIPIENT_ADD = 1
    RECIPIENT_REMOVE = 2
    CALL = 3
    CHANNEL_NAME_CHANGE = 4
    CHANNEL_ICON_CHANGE = 5
    CHANNEL_PINNED_MESSAGE = 6
    GUILD_MEMBER_JOIN = 7
    USER_PREMIUM_GUILD_SUBSCRIPTION = 8
    USER_PREMIUM_GUILD_SUBSCRIPTION_TIER_1 = 9
    USER_PREMIUM_GUILD_SUBSCRIPTION_TIER_2 = 10
    USER_PREMIUM_GUILD_SUBSCRIPTION_TIER_3 = 11
    CHANNEL_FOLLOW_ADD = 12
    GUILD_DISCOVERY_DISQUALIFIED = 14
    GUILD_DISCOVERY_REQUALIFIED = 15
    GUILD_DISCOVERY_GRACE_PERIOD_INITIAL_WARNING = 16
    GUILD_DISCOVERY_GRACE_PERIOD_FINAL_WARNING = 17
    THREAD_CREATED = 18
    REPLY = 19
    CHAT_INPUT_COMMAND = 20
    THREAD_STARTER_MESSAGE = 21
    GUILD_INVITE_REMINDER = 22
    CONTEXT_MENU_COMMAND = 23


class CHANNEL_TYPES(enum.IntEnum):
    '''Types of channels.'''
    GUILD_TEXT = 0  # a text channel within a server
    DM = 1  # a direct message between users
    GUILD_VOICE = 2  # a voice channel within a server
    GROUP_DM = 3  # a direct message between multiple users
    GUILD_CATEGORY = 4  # an organizational category that contains up to 50 channels
    GUILD_NEWS = 5  # a channel that users can follow and crosspost into their own server
    GUILD_STORE = 6  # a channel in which game developers can sell their game on Discord
    GUILD_NEWS_THREAD = 10  # a temporary sub-channel within a GUILD_NEWS channel
    GUILD_PUBLIC_THREAD = 11  # a temporary sub-channel within a GUILD_TEXT channel
    GUILD_PRIVATE_THREAD = 12  # a temporary sub-channel within a GUILD_TEXT channel that is only viewable by those invited and those with the MANAGE_THREADS permission
    GUILD_STAGE_VOICE = 13  # a voice channel for hosting events with an audience


class EMBED_TYPES(enum.Enum):
    '''Types of Embeds.'''
    rich = enum.auto()      # generic embed rendered from embed attributes
    image = enum.auto()     # image embed
    video = enum.auto()     # video embed
    gifv = enum.auto()      # animated gif image embed rendered as a video embed
    article = enum.auto()   # article embed
    link = enum.auto()      # link embed



import enum
import datetime
from typing import List

from . import snowflake, role
from .base_object import BaseDiscordObject
from .. import utilities


class User(BaseDiscordObject):
    '''Discord User.'''

    @enum.unique
    class UserFlag(enum.IntFlag):
        '''Flags on a User.'''
        NONE = 0
        DISCORD_EMPLOYEE = 1 << 0
        PARTNERED_SERVER_OWNER = 1 << 1
        HYPESQUAD_EVENTS = 1 << 2
        BUG_HUNTER_LEVEL_1 = 1 << 3
        HOUSE_BRAVERY = 1 << 6
        HOUSE_BRILLIANCE = 1 << 7
        HOUSE_BALANCE = 1 << 8
        EARLY_SUPPORTER = 1 << 9
        TEAM_USER = 1 << 10
        BUG_HUNTER_LEVEL_2 = 1 << 14
        VERIFIED_BOT = 1 << 16
        EARLY_VERIFIED_BOT_DEVELOPER = 1 << 17
        DISCORD_CERTIFIED_MODERATOR = 1 << 18

    @enum.unique
    class PremiumType(enum.IntFlag):
        '''Type of Premium subscription the User has.'''
        NONE = 0
        NITRO_CLASSIC = 1 << 0
        NITRO = 1 << 1

    id: snowflake.Snowflake
    username: str
    discriminator: str
    avatar: str
    bot: bool
    system: bool
    mfa_enabled: bool
    banner: str
    accent_color: int  # TODO: This is actually a color object!
    locale: str
    verified: bool
    email: str
    flags: int
    premium_type: int
    public_flags: int

    def __str__(self):
        '''Return string representation.'''
        fields = []

        if 'id' in self.__dict__:
            fields.append(f'id={self.id}')

        if ('username' in self.__dict__) and ('discriminator' in self.__dict__):
            fields.append(f'username=\'{self.username}#{self.discriminator}\'')

        return f'User({", ".join(fields)})'

    def __repr__(self):
        '''Return string representation.'''
        return self.__str__()

    def __eq__(self, other):
        '''Determine if other User and self are the same User.'''
        if not isinstance(other, User):
            return NotImplemented
        return self.id == other.id

    # TODO: We could have a lot of handy "is_X" functions here to return if user is something form the USER_FLAG enumeration

    @property
    def mention(self):
        '''Generate a valid name mention for use in Discord.'''
        return f'<@{self.id}>'

    @property
    def mention_nickname(self):
        '''Generate a valid nickname mention for use in Discord.'''
        return f'<@!{self.id}>'

    def ingest_raw_dict(self, data: dict) -> 'User':
        '''Parse a User from an API compliant dict.'''
        self.from_dict(data)

        self.cache()

        return self

    def from_dict(self, data: dict) -> 'User':
        '''Parse a User from an API compliant dict.'''
        # Required fields
        self.id = snowflake.Snowflake(data['id'])
        self.username = data['username']
        self.discriminator = data['discriminator']
        self.avatar = data['avatar']

        # Optional fields
        if 'bot' in data:
            self.bot = data['bot']
        if 'system' in data:
            self.system = data['system']

        return self

    def cache(self):
        '''Deprecated caching method.'''
        utilities.Cache().add(self)


class Member(User):
    '''User within a guild.'''

    nick: str
    avatar: str
    roles: List[role.Role]
    joined_at: datetime.datetime
    premium_since: datetime.datetime
    deaf: bool
    mute: bool
    pending: bool
    permissions: str

    def __str__(self):
        '''Return string representation.'''
        fields = []

        if hasattr(self, 'id'):
            fields.append(f'id={self.id}')

        if hasattr(self, 'username') and hasattr(self, 'discriminator'):
            fields.append(f'username=\'{self.username}#{self.discriminator}\'')

        if hasattr(self, 'nick'):
            fields.append(f'nick=\'{self.nick}\'')

        return f'Member({", ".join(fields)})'

    def from_dict(self, data: dict) -> 'Member':
        '''Parse a Member from an API compliant dict.'''
        if 'nick' in data:
            self.nick = data['nick']
        if 'avatar' in data:
            self.avatar = data['avatar']
        if 'joined_at' in data:
            self.joined_at = datetime.datetime.fromisoformat(data['joined_at'])
        if 'premium_since' in data and data['premium_since'] is not None:
            self.premium_since = datetime.datetime.fromisoformat(data['premium_since'])
        if 'deaf' in data:
            self.deaf = data['deaf']
        if 'mute' in data:
            self.mute = data['mute']
        if 'pending' in data:
            self.pending = data['pending']
        if 'permissions' in data:
            self.permissions = data['permissions']
        if 'user' in data:
            tmp_user = User().from_dict(data['user'])
            self.update_from_user(tmp_user)
        return self

    def update_from_user(self, user: User) -> 'Member':  # noqa: C901
        '''Update Member properties from associated user.'''
        if hasattr(user, 'id'):
            self.id = user.id
        if hasattr(user, 'username'):
            self.username = user.username
        if hasattr(user, 'discriminator'):
            self.discriminator = user.discriminator
        if hasattr(user, 'avatar'):
            self.avatar = user.avatar
        if hasattr(user, 'bot'):
            self.bot = user.bot
        if hasattr(user, 'system'):
            self.system = user.system
        if hasattr(user, 'mfa_enabled'):
            self.mfa_enabled = user.mfa_enabled
        if hasattr(user, 'banner'):
            self.banner = user.banner
        if hasattr(user, 'accent_color'):
            self.accent_color = user.accent_color
        if hasattr(user, 'locale'):
            self.locale = user.locale
        if hasattr(user, 'verified'):
            self.verified = user.verified
        if hasattr(user, 'email'):
            self.email = user.email
        if hasattr(user, 'flags'):
            self.flags = user.flags
        if hasattr(user, 'premium_type'):
            self.premium_type = user.premium_type
        if hasattr(user, 'public_flags'):
            self.public_flags = user.public_flags
        return self

    def ingest_raw_dict(self, data: dict) -> 'User':
        '''Ingest and cache a given object for future use.'''
        self.from_dict(data)

        self.cache()

        return self

    def cache(self):
        '''Deprecated caching function.'''
        utilities.Cache().add(self)



'''Dyscord is a library to enable a reasonably 1:1 binding between python and the official API.'''

from . import client, helper, objects, utilities

__all__ = [
    'client',
    'helper',
    'objects',
    'utilities',
]



import asyncio
import datetime
import enum
from typing import Union, Optional, List, Dict

from .base_object import BaseDiscordObject
from . import user, channel as ext_channel, snowflake, enumerations, embed as ext_embed, guild as ext_guild, role as ext_role
from ..utilities import cache, log
from .interactions import components as ext_components

from ..client import api


class Message(BaseDiscordObject, ext_components.ComponentAdder, ext_embed.EmbedAdder):
    '''Message containing infomation about it's content, origin, authors, etc.'''

    _log = log.Log()

    MESSAGE_TYPE = enumerations.MESSAGE_TYPE

    id: 'snowflake.Snowflake' = None  # type: ignore
    channel_id: 'snowflake.Snowflake' = None  # type: ignore
    guild_id: Optional['snowflake.Snowflake'] = None  # type: ignore
    guild: Optional['ext_guild.Guild'] = None  # type: ignore
    author: 'user.User' = None  # type: ignore
    member: Optional['user.Member'] = None  # type: ignore
    content: str = None  # type: ignore
    timestamp: datetime.datetime = None  # type: ignore
    edited_timestamp: Optional[datetime.datetime] = None  # type: ignore
    tts: bool = None  # type: ignore
    mention_everyone: bool = None  # type: ignore
    mentions: List['user.User'] = None  # type: ignore
    mention_roles: List['ext_role.Role'] = None  # type: ignore # Roles are TBD
    mention_channels: List['ext_channel.Channel'] = None  # type: ignore
    attachments: 'List' = None  # type: ignore # attachment.Attachment TDB
    embeds: Optional[List[ext_embed.Embed]] = None  # type: ignore
    # reactions: 'List[reaction.Reaction]'  = None  # type: ignore
    nonce: Optional[Union[int, str]] = None  # type: ignore
    pinned: bool = None  # type: ignore
    webhook_id: Optional['snowflake.Snowflake'] = None  # type: ignore
    type: 'enumerations.MESSAGE_TYPE' = None  # type: ignore
    # activity = None  # Activity TBD = None  # type: ignore
    # application: application.Application = None  # Application TBD = None  # type: ignore
    application_id: 'snowflake.Snowflake' = None  # type: ignore
    # message_reference: 'MessageReference' = None  # Message Reference TBD = None  # type: ignore
    flags: int  # Message Flags = None  # type: ignore
    referenced_message: 'Message' = None  # type: ignore
    # interaction = None  # Interactions TBD = None  # type: ignore
    thread: 'ext_channel.Channel' = None  # type: ignore
    components: List['ext_components.Component']  # Message components TBD = None  # type: ignore
    # sticker_items: List[Sticker] = []  # Stickers TBD = None  # type: ignore

    def __init__(self,
                 content: str = None
                 ):
        '''Init.'''
        if content is not None:
            self.content = content

    def __getattr__(self, name):
        '''Do dynamic lookup on various fields that may not be populated, but have valid representations in the API.'''
        if name == 'channel' and 'channel_id' in self.__dict__:
            try:
                channel = cache.Cache().get(self.channel_id)
                self._log.info('Got channel from the cache.')
            except LookupError:
                loop = asyncio.get_event_loop()
                channel_dict = loop.run_until_complete(api.API.get_channel(self.channel_id))
                channel = ext_channel.ChannelImporter().ingest_raw_dict(channel_dict)
                self._log.info('Got channel from the API.')
            self.channel = channel
            return channel

        if name == 'guild' and 'guild_id' in self.__dict__:
            assert type(self.guild_id) is snowflake.Snowflake
            try:
                guild = cache.Cache().get(self.guild_id)
                self._log.info('Got guild from the cache.')
            except LookupError:
                loop = asyncio.get_event_loop()
                guild_dict = loop.run_until_complete(api.API.get_guild(self.guild_id))
                guild = ext_guild.Guild().from_dict(guild_dict)
                self._log.info('Got guild from the API.')
            self.guild = guild
            return guild
        raise AttributeError(f'Failed to find \'{name}\'')

    def ingest_raw_dict(self, data) -> 'Message':
        '''Ingest and cache a given object for future use.'''
        self.from_dict(data)

        # I do not believe we need to cache these at this time.
        # self.cache()
        return self

    def from_dict(self, data: dict) -> 'Message':
        '''Parse a Message from an API compliant dict.'''
        # Mandatory Fields
        self.attachments = list()
        if 'attachments' in data:
            for attachment in data['attachments']:
                self.attachments.append(attachment)
        self.author = user.User().from_dict(data['author'])  # TODO: Update after we can parse in users.
        self.channel_id = snowflake.Snowflake(data['channel_id'])
        self.components = list()
        for component in data['components']:
            self.components.append(component)
        self.content = data['content']
        self.edited_timestamp = datetime.datetime.fromisoformat(data['edited_timestamp']) if data['edited_timestamp'] is not None else None
        self.embeds = list()
        for embed in data['embeds']:
            self.embeds.append(embed)
        self.flags = data['flags']
        self.id = snowflake.Snowflake(data['id'])
        self.mention_everyone = data['mention_everyone']
        self.mention_roles = []
        for role in data['mention_roles']:
            # self.mention_roles.append(role.Role().from_dict(role))
            pass
        self.mentions = []
        for user_dict in data['mentions']:
            self.mentions.append(user.User().from_dict(user_dict))
        self.pinned = data['pinned']
        self.timestamp = datetime.datetime.fromisoformat(data['timestamp'])
        self.tts = data['tts']
        self.type = enumerations.MESSAGE_TYPE(data['type'])

        # Optional fields
        self.guild_id = snowflake.Snowflake(data['guild_id']) if (data.get('guild_id', None) is not None) else None
        if 'member' in data:
            self.member = user.Member().from_dict(data['member'])  # TODO: Update after we can parse in users.
            self.member.update_from_user(self.author)
        self.nonce = data['nonce'] if 'nonce' in data else None
        self.webhook_id = snowflake.Snowflake(data['webhook_id']) if (data.get('webhook_id', None) is not None) else None

        return self

    def cache(self):
        '''Sav object to the cache for faster recall in the future.'''
        raise NotImplementedError(f'{__class__.__name__} does not yet implement this function.')

    def to_sendable_dict(self) -> dict:
        '''Sending a message only allows a subset of attributes. Ignore anything else about this message when producing that dict.'''
        new_dict: Dict[str, object] = dict()
        if self.content is not None:
            new_dict['content'] = self.content
        new_dict['tts'] = self.tts if hasattr(self, 'tts') else False
        # new_dict['file'] = None  # TODO: Handle a file upload.
        if self.embeds is not None:
            new_dict['embeds'] = list()  # TODO: Handle embeds.
            assert type(new_dict['embeds']) is list
            assert type(self.embeds) is list
            for embed in self.embeds:
                new_dict['embeds'].append(embed.to_dict())
        # new_dict['allowed_mentions'] = True  # BUG: This isn't a bool, its an `AllowedMentionsObject`, which we don't support yet.
        # new_dict['message_reference'] = None
        # new_dict['sticker_ids'] = None

        if isinstance(self.components, list) and len(self.components):
            new_dict['components'] = list()
            assert type(new_dict['components']) is list
            for component in self.components:
                new_dict['components'].append(component.to_dict())

        return new_dict

    def validate(self):
        '''Validate object is prepared for dispatch to discord.'''
        assert (self.content is not None) or (self.embeds is not None) or (self.attachments is not None),\
            'Message must have a content, embeds, or attachments to be valid for sending.'

        if self.content is not None:
            assert type(self.content) is str,\
                f'Got invalid type {type(self.content)} in Message.content, must be str.'

            assert len(self.content) <= 2000,\
                f'Invalid length of {len(self.content):,} in Message.content, must be <= 2000 characters.'

        if self.components is not None:
            assert type(self.components) is list,\
                f'Got invalid type {type(self.components)} in Message.components, must be list.'

            assert len(self.components) <= 5,\
                f'Invalid length of {len(self.components):,} in Message.components, must be <= 5 elements.'

            custom_ids = list()
            for component in self.components:
                assert type(component) is ext_components.ActionRow,\
                    f'Got invalid type {type(component)} in Message.components, must be ActionRow.'
                component.validate()
                for sub_component in component.components:
                    if sub_component.custom_id is not None:
                        assert sub_component.custom_id not in custom_ids,\
                            f'Found duplicate custom_id [{sub_component.custom_id}]'
                        custom_ids.append(sub_component.custom_id)

        if self.embeds is not None:
            assert type(self.embeds) is list,\
                f'Got invalid type {type(self.embeds)} in Message.embeds, must be list.'

            assert len(self.embeds) <= 25,\
                f'Invalid length of {len(self.embeds):,} in Message.embeds, must be <= 5 elements.'

            for embed in self.embeds:
                embed.validate()

    class formatter:
        '''Helper formatting functions.'''

        class TIMESTAMP_FLAGS(enum.Enum):
            '''Types of timestamp displays.'''
            SHORT_TIME = 't'
            LONG_TIME = 'T'
            SHORT_DATE = 'd'
            LONG_DATE = 'D'
            DEFAULT = 'f'
            LONG_DATE_TIME = 'F'
            RELATIVE_TIME = 'R'

        @classmethod
        def timestamp(cls, timestamp: datetime.datetime, flag: Union[TIMESTAMP_FLAGS, str] = TIMESTAMP_FLAGS.DEFAULT) -> str:
            '''Return valid timestamp string. Discord will display this in a useful way.'''
            if type(flag) is cls.TIMESTAMP_FLAGS:
                flag = flag.value
            return f'<t:{int(timestamp.timestamp())}:{flag}>'

        @classmethod
        def user(cls, user_id: snowflake.Snowflake):
            '''Return valid mention string for the user's name.'''
            return f'<@{user_id}>'

        @classmethod
        def user_nickname(cls, user_id: snowflake.Snowflake):
            '''Return valid mention string for the user's nickname.'''
            return f'<@!{user_id}>'


class MessageReference(BaseDiscordObject):
    '''Slim data holding class.'''

    message_id: Optional[snowflake.Snowflake] = None  # id of the originating message
    channel_id: Optional[snowflake.Snowflake] = None  # id of the originating message's channel
    guild_id: Optional[snowflake.Snowflake] = None  # id of the originating message's guild
    fail_if_not_exists: Optional[bool] = None  # when sending, whether to error if the referenced message doesn't exist instead of sending as a normal (non-reply) message, default true
